import os

import requests
import yaml

_ARTIFACT_NAME = 'bubuku-appliance'


class ConfigLoader():
    def load_config(self) -> dict:
        pass


class AwsInstanceUserDataLoader(ConfigLoader):
    def load_config(self):
        return yaml.load(requests.get('http://169.254.169.254/latest/user-data').text)


class ClusterConfig():

    def __init__(self, config_loader: ConfigLoader):
        self._user_data = config_loader.load_config()
        self._env_vars = self._user_data.get('environment')

    def get_cluster_name(self):
        return self._env_vars.get('CLUSTER_NAME')

    def get_health_port(self):
        return self._env_vars.get('HEALTH_PORT')

    def get_aws_region(self):
        return 'eu-central-1'

    def get_instance_type(self):
        instance_type = self._user_data.get('instance_type')
        if instance_type is None:
            instance_type = os.getenv('INSTANCE_TYPE', 'm4.4xlarge')
            self.set_instance_type(instance_type)
        return instance_type

    def set_scalyr_account_key(self, scalyr_account_key):
        if scalyr_account_key:
            self._user_data['scalyr_account_key'] = scalyr_account_key

    def set_scalyr_region(self, scalyr_region):
        if scalyr_region:
            self._user_data['scalyr_region'] = scalyr_region

    def set_application_version(self, image):
        if image:
            self._user_data['application_version'] = image
            self._user_data['source'] = 'registry.opensource.zalan.do/aruha/{}:{}'.format(_ARTIFACT_NAME, image)

    def get_application_version(self):
        return self._user_data.get('application_version')

    def set_instance_type(self, instance_type):
        if instance_type:
            self._user_data['instance_type'] = instance_type

    def set_kms_key_id(self, kms_key_id):
        if kms_key_id:
            self._user_data['kms_key_id'] = kms_key_id

    def get_kms_key_id(self):
        return self._user_data.get("kms_key_id")

    def get_availability_zone(self):
        return self._user_data.get('availability_zone')

    def set_availability_zone(self, availability_zone):
        if availability_zone:
            self._user_data['availability_zone'] = availability_zone

    def should_create_ebs(self):
        return self._user_data.get('should_create_ebs')

    def set_should_create_ebs(self, v):
        self._user_data['should_create_ebs'] = v

    def get_volume_size(self):
        return self._user_data.get('volume_size')

    def get_volume_type(self):
        return self._user_data.get('volume_type')

    def get_vpc_id(self):
        vpc_id = self._user_data.get('vpc_id')
        if vpc_id is None:
            vpc_id = os.getenv('VPC_ID')
            self._user_data['vpc_id'] = vpc_id
        return vpc_id

    def get_user_data(self):
        return dict(self._user_data)



import copy
import logging
import time

import yaml

from bubuku.aws import volume, AWSResources, taupage, metric, subnet, security_group, iam
from bubuku.aws.cluster_config import ClusterConfig

_LOG = logging.getLogger('bubuku.aws.ec2_node')


class EC2(object):
    def __init__(self, aws: AWSResources):
        self.aws = aws

    def _create_tagged_volume(self, cluster_config: ClusterConfig, zone: str, name: str):
        _LOG.info('Creating EBS volume %s in %s', name, zone)
        vol = self.aws.ec2_client.create_volume(
            AvailabilityZone=zone,
            VolumeType=cluster_config.get_volume_type(),
            Size=cluster_config.get_volume_size(),
            Encrypted=False,
        )
        _LOG.info('%s is successfully created', vol['VolumeId'])

        _LOG.info('Tagging %s with Taupage:erase-on-boot, to format only once', vol['VolumeId'])
        tags = [
            {
                'Key': 'Name',
                'Value': name
            }, {
                'Key': 'Taupage:erase-on-boot',
                'Value': 'True'
            }
        ]
        self.aws.ec2_client.create_tags(Resources=[vol['VolumeId']], Tags=tags)

    def _launch_instance(self, cluster_config: ClusterConfig, ip: str, subnet_: dict, ami: object,
                         security_group_id: str, iam_profile):
        _LOG.info('Launching node %s in %s', ip, subnet_['AvailabilityZone'])

        #
        # Override any ephemeral volumes with NoDevice mapping,
        # otherwise auto-recovery alarm cannot be actually enabled.
        #
        _LOG.info('Overriding ephemeral volumes to be able to set up AWS auto recovery alarm ')
        block_devices = []
        for bd in ami.block_device_mappings:
            if 'Ebs' in bd:
                #
                # This has to be our root EBS.
                #
                # If the Encrypted flag is present, we have to delete
                # it even if it matches the actual snapshot setting,
                # otherwise amazon will complain rather loudly.
                #
                # Take a deep copy before deleting the key:
                #
                bd = copy.deepcopy(bd)

                root_ebs = bd['Ebs']
                if 'Encrypted' in root_ebs:
                    del (root_ebs['Encrypted'])

                block_devices.append(bd)
            else:
                # ignore any ephemeral volumes (aka. instance storage)
                block_devices.append({
                    'DeviceName': bd['DeviceName'],
                    'NoDevice': ''
                })

        if cluster_config.should_create_ebs():
            self._create_tagged_volume(cluster_config, subnet_['AvailabilityZone'], volume.KAFKA_LOGS_EBS)

        user_data = cluster_config.get_user_data()
        user_data['volumes']['ebs']['/dev/xvdk'] = volume.KAFKA_LOGS_EBS
        taupage_user_data = '#taupage-ami-config\n{}'.format(yaml.safe_dump(user_data))

        resp = self.aws.ec2_client.run_instances(
            ImageId=ami.id,
            MinCount=1,
            MaxCount=1,
            SecurityGroupIds=[security_group_id],
            UserData=taupage_user_data,
            InstanceType=cluster_config.get_instance_type(),
            SubnetId=subnet_['SubnetId'],
            PrivateIpAddress=ip,
            BlockDeviceMappings=block_devices,
            IamInstanceProfile={'Arn': iam_profile['Arn']},
            DisableApiTermination=False,
            EbsOptimized=True)

        instance_id = resp['Instances'][0]['InstanceId']
        _LOG.info('Instance %s launched, waiting for it to initialize', instance_id)
        self.aws.ec2_client.create_tags(
            Resources=[instance_id],
            Tags=[
                {'Key': 'Name', 'Value': cluster_config.get_cluster_name()},
                {'Key': 'StackName', 'Value': cluster_config.get_cluster_name()}
            ]
        )

        return instance_id

    def _launch_nodes(self, cluster_config: ClusterConfig, node_ips: list, taupage_amis, security_groups, iam_profile):
        starting_instances = []
        for subnet_, ip in node_ips:
            starting_instances.append(
                self._launch_instance(
                    cluster_config,
                    ip,
                    subnet_,
                    taupage_amis,
                    security_groups['GroupId'],
                    iam_profile))
            # wait for all instances to start
        while starting_instances:
            _LOG.info("Waiting for instances to start: {}".format(starting_instances))
            time.sleep(5)
            resp = self.aws.ec2_client.describe_instances(InstanceIds=starting_instances)
            started_instances = []
            for r in resp['Reservations']:
                started_instances += [i['InstanceId'] for i in r['Instances'] if i['State']['Name'] != 'pending']
            if started_instances:
                _LOG.info('Instances {} started'.format(started_instances))
            for instance_id in started_instances:
                starting_instances.remove(instance_id)
                metric.create_auto_recovery_alarm(self.aws, cluster_config.get_cluster_name(), instance_id)

    def create(self, cluster_config: ClusterConfig, instance_count: int):
        _LOG.info('Preparing AWS configuration for ec2 instance creation')
        try:
            node_ips = subnet.allocate_ip_addresses(self.aws, cluster_config, instance_count)
            taupage_amis = taupage.find_amis(self.aws.ec2_resource, cluster_config.get_aws_region())
            security_groups = security_group.create_or_ger_security_group(self.aws, cluster_config)
            iam_profile = iam.create_or_get_instance_profile(self.aws, cluster_config)

            self._launch_nodes(cluster_config, node_ips, taupage_amis, security_groups, iam_profile)

        except Exception as e:
            _LOG.error('''
                    You were trying to deploy Bubuku, but the process has failed :-(
                    
                    One of the reasons might be that some of Private IP addresses we were
                    going to use to launch the EC2 instances were taken by some other
                    instances in the middle of the process.  If that is the case, simply
                    retrying the operation might resolve the problem (you still might need
                    to clean up after this attempt before retrying).
                    
                    Please review the error message to see if that is the case, then
                    either correct the error or retry.
                    
                ''')
            raise e



from subprocess import call

from instance_control import config


class Command(object):
    def __init__(self, cluster_config_path: str):
        self.cluster_config = config.read_cluster_config(cluster_config_path)

    def alter_config(self):
        pass

    def execute(self):
        raise NotImplementedError('Not implemented yet')

    def run(self):
        self.alter_config()
        config.validate_config(self.cluster_config)
        call(["zaws", "login", self.cluster_config['account']])
        self.execute()



import logging
from time import time

import requests

from bubuku.api import ApiConfig
from bubuku.aws import AWSResources, node, volume
from bubuku.aws.cluster_config import ClusterConfig
from bubuku.aws.ec2_node import EC2
from bubuku.controller import Change
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.features.rolling_restart')


class RollingRestartChange(Change):
    def __init__(self, zk: BukuExhibitor, cluster_config: ClusterConfig,
                 broker_id_to_restart: str,
                 image: str,
                 instance_type: str,
                 scalyr_key: str,
                 scalyr_region: str,
                 kms_key_id: str):
        self.zk = zk
        self.cluster_config = cluster_config
        self.cluster_config.set_application_version(image)
        self.cluster_config.set_instance_type(instance_type)
        self.cluster_config.set_scalyr_account_key(scalyr_key)
        self.cluster_config.set_scalyr_region(scalyr_region)
        self.cluster_config.set_kms_key_id(kms_key_id)
        self.state_context = StateContext(self.cluster_config, broker_id_to_restart,
                                          self.zk.get_broker_address(broker_id_to_restart))

    def get_name(self) -> str:
        return 'rolling_restart'

    def can_run(self, current_actions):
        return all([a not in current_actions for a in ['start', 'stop', 'restart', 'rebalance']])

    def run(self, current_actions) -> bool:
        if not self._is_cluster_in_good_state():
            _LOG.error('Cluster is not stable skipping restart iteration')
            return True

        return self.state_context.run()

    def _is_cluster_in_good_state(self):
        return True


class StateContext:
    def __init__(self, cluster_config, broker_id_to_restart, broker_ip_to_restart):
        self.broker_id_to_restart = broker_id_to_restart
        self.broker_ip_to_restart = broker_ip_to_restart
        self.cluster_config = cluster_config
        self.aws = AWSResources(region=self.cluster_config.get_aws_region())
        self.current_state = StopKafka(self)
        self.state_wait_broker_stopped = WaitBrokerStopped(self)
        self.state_detach_volume = DetachVolume(self)
        self.state_terminate_instance = TerminateInstance(self)
        self.state_launch_instance = LaunchInstance(self)
        self.state_volume_attached = WaitVolumeAttach(self)

    def run(self):
        """
        Runs states one after another. If state is finished, it takes the next one.
        """
        try:
            if self.current_state.run():
                next_state = self.current_state.next()
                _LOG.info('Next state is {}'.format(next_state))
                if next_state is None:
                    return False
                self.current_state = next_state
            return True
        except Exception as e:
            _LOG.error('Failed to run state', exc_info=e)
            return True


class State:
    """
    State which can be run as many times as required before it finishes it work. The progress of the state has to be
    recoverable
    """

    def __init__(self, state_context):
        self.state_context = state_context
        self.time_to_check_s = time()

    def run(self) -> bool:
        """
        Runs the state, and if state finishes successfully it returns True, otherwise it returns False, which means
        that state has to be executed again
        """
        pass

    def next(self):
        """
        Return the next state, which has to be executed after the current state
        """
        pass

    def run_with_timeout(self, func, timeout_s=10):
        """
        Runs func() with timeout
        :param func function to execute
        :param timeout_s timeout before executing state next time
        """
        if time() >= self.time_to_check_s:
            self.time_to_check_s = time() + timeout_s
            return func()
        return False


class StopKafka(State):
    def run(self):
        _LOG.info('Stopping broker {} {}'.format(self.state_context.broker_id_to_restart,
                                                 self.state_context.broker_ip_to_restart))
        resp = requests.post(ApiConfig.get_url(self.state_context.broker_ip_to_restart, 'stop'))
        if resp.status_code != 200:
            _LOG.error('Failed to stop Kafka: {} {}', resp.status_code, resp.text)
            return False
        return True

    def next(self):
        return self.state_context.state_wait_broker_stopped


class WaitBrokerStopped(State):
    def __init__(self, state_context):
        super(WaitBrokerStopped, self).__init__(state_context)
        self.time_to_check_s = time()

    def run(self):
        def func():
            resp = requests.get(ApiConfig.get_url(self.state_context.broker_ip_to_restart, 'state')).json()
            return resp.get('state') == 'stopped'

        return self.run_with_timeout(func)

    def next(self):
        return self.state_context.state_detach_volume


class DetachVolume(State):
    def run(self):
        instance = node.get_instance_by_ip(self.state_context.aws.ec2_resource,
                                           self.state_context.cluster_config,
                                           self.state_context.broker_ip_to_restart)

        _LOG.info('Searching for instance %s volumes', instance.instance_id)
        volumes = self.state_context.aws.ec2_client.describe_instance_attribute(InstanceId=instance.instance_id,
                                                                                Attribute='blockDeviceMapping')
        data_volume = next(v for v in volumes['BlockDeviceMappings'] if v['DeviceName'] == '/dev/xvdk')
        data_volume_id = data_volume['Ebs']['VolumeId']

        _LOG.info('Creating tag:Name=%s for %s', volume.KAFKA_LOGS_EBS, data_volume_id)
        vol = self.state_context.aws.ec2_resource.Volume(data_volume_id)
        self.state_context.cluster_config.set_availability_zone(vol.availability_zone)
        vol.create_tags(Tags=[{'Key': 'Name', 'Value': volume.KAFKA_LOGS_EBS}])
        _LOG.info('Detaching %s from %s', data_volume_id, instance.instance_id)

        self.state_context.aws.ec2_client.detach_volume(VolumeId=data_volume_id, Force=False)
        return True

    def next(self):
        return self.state_context.state_terminate_instance


class TerminateInstance(State):
    def __init__(self, state_context):
        super(TerminateInstance, self).__init__(state_context)
        self.instance = None

    def run(self):
        if self.instance is None:
            self.instance = node.get_instance_by_ip(self.state_context.aws.ec2_resource,
                                                    self.state_context.cluster_config,
                                                    self.state_context.broker_ip_to_restart)

        def func():
            node.terminate(self.state_context.aws, self.state_context.cluster_config, self.instance)
            return True

        return self.run_with_timeout(func)

    def next(self):
        return self.state_context.state_launch_instance


class LaunchInstance(State):
    def run(self):
        ec2 = EC2(self.state_context.aws)
        ec2.create(self.state_context.cluster_config, 1)
        return True

    def next(self):
        return self.state_context.state_volume_attached


class WaitVolumeAttach(State):
    def run(self):
        def func():
            return volume.are_volumes_attached(self.state_context.aws)

        return self.run_with_timeout(func)

    def next(self):
        return None



import logging

import click
import requests
from requests import Response

from bubuku.features.remote_exec import RemoteCommandExecutorCheck
from bubuku.utils import get_opt_broker_id, prepare_configs
from bubuku.zookeeper import load_exhibitor_proxy, BukuExhibitor

_LOG = logging.getLogger('bubuku.cli')


def _print_table(table: list, print_function=None):
    if not print_function:
        print_function = print
    names = sorted(set([v for v in sum([list(k.keys()) for k in table], [])]))
    lengths = {n: len(n) for n in names}
    for d in table:
        for k, v in d.items():
            if lengths[k] < len(str(v)):
                lengths[k] = len(str(v))
    format_string = '  '.join(['{!s:' + str(lengths[n]) + 's}' for n in names])
    print_function(format_string.format(*names))
    for item in table:
        print_function(format_string.format(*[item.get(n, '') for n in names]))


def __validate_not_empty(ctx, param, value):
    if not value:
        raise click.BadParameter('Parameter must have value')
    return value


def __check_all_broker_ids_exist(broker_ids: list, zk: BukuExhibitor):
    registered_brokers = zk.get_broker_ids()
    unknown_brokers = [broker_id for broker_id in broker_ids if broker_id not in registered_brokers]
    if len(unknown_brokers) == 1:
        raise Exception('1 broker id is not valid: {}'.format(unknown_brokers[0]))
    if len(unknown_brokers) > 1:
        raise Exception('{} broker ids are not valid: {}'.format(len(unknown_brokers), ",".join(unknown_brokers)))


logging.basicConfig(level=getattr(logging, 'INFO', None))


@click.group()
def cli():
    logo = """
        ____        __          __        
       / __ )__  __/ /_  __  __/ /____  __
      / __  / / / / __ \/ / / / //_/ / / /
     / /_/ / /_/ / /_/ / /_/ / ,< / /_/ / 
    /_____/\__,_/_.___/\__,_/_/|_|\__,_/  
    """
    print(logo)
    print('Start, monitor and rebalance kafka cluster in AWS setup')
    print()
    pass


@cli.command('restart', help='Restart kafka instance')
@click.option('--broker', type=click.STRING,
              help='Broker id to restart. By default current broker id is restarted')
def restart_broker(broker: str):
    config, env_provider = prepare_configs()
    with load_exhibitor_proxy(env_provider.get_address_provider(), config.zk_prefix) as zookeeper:
        broker_id = get_opt_broker_id(broker, config, zookeeper, env_provider)
        RemoteCommandExecutorCheck.register_restart(zookeeper, broker_id)


@cli.command('rolling-restart', help='Rolling restart of Kafka cluster')
@click.option('--image-tag', type=click.STRING, help='Docker image to run Kafka broker')
@click.option('--instance-type', type=click.STRING, default='m4.4xlarge',
              help='AWS instance type to run Kafka broker on')
@click.option('--scalyr-key', type=click.STRING, help='Scalyr account key')
@click.option('--scalyr-region', type=click.STRING, help='Scalyr region to use')
@click.option('--kms-key-id', type=click.STRING, help='Kms key id to decrypt data with')
def rolling_restart_broker(image_tag: str, instance_type: str, scalyr_key: str, scalyr_region: str, kms_key_id: str):
    config, env_provider = prepare_configs()
    with load_exhibitor_proxy(env_provider.get_address_provider(), config.zk_prefix) as zookeeper:
        broker_id = get_opt_broker_id(None, config, zookeeper, env_provider)
        RemoteCommandExecutorCheck.register_rolling_restart(zookeeper, broker_id, image_tag, instance_type, scalyr_key,
                                                            scalyr_region, kms_key_id)


@cli.command('rebalance', help='Run rebalance process on one of brokers. If rack-awareness is enabled, replicas will '
                               'only be move to other brokers in the same rack')
@click.option('--broker', type=click.STRING,
              help="Broker instance on which to perform rebalance. By default, any free broker will start it")
@click.option('--empty_brokers', type=click.STRING,
              help="Comma-separated list of brokers to empty. All partitions will be moved to other brokers")
@click.option('--exclude_topics', type=click.STRING, help="Comma-separated list of topics to exclude from rebalance")
@click.option('--bin-packing', is_flag=True, help="Use bean packing approach instead of one way processing")
@click.option('--parallelism', type=click.INT, default=1, show_default=True,
              help="Amount of partitions to move in a single rebalance step")
def rebalance_partitions(broker: str, empty_brokers: str, exclude_topics: str, parallelism: int, bin_packing: bool):
    config, env_provider = prepare_configs()
    with load_exhibitor_proxy(env_provider.get_address_provider(), config.zk_prefix) as zookeeper:
        empty_brokers_list = [] if empty_brokers is None else empty_brokers.split(',')
        exclude_topics_list = [] if exclude_topics is None else exclude_topics.split(',')
        __check_all_broker_ids_exist(empty_brokers_list, zookeeper)
        broker_id = get_opt_broker_id(broker, config, zookeeper, env_provider) if broker else None
        RemoteCommandExecutorCheck.register_rebalance(zookeeper, broker_id, empty_brokers_list,
                                                      exclude_topics_list, parallelism, bin_packing)


@cli.command('migrate', help='Replace one broker with another for all partitions')
@click.option('--from', 'from_', type=click.STRING, callback=__validate_not_empty,
              help='List of brokers to migrate from (separated with ",")')
@click.option('--to', type=click.STRING, callback=__validate_not_empty,
              help='List of brokers to migrate to (separated with ",")')
@click.option('--shrink', is_flag=True, default=False, show_default=True,
              help='Whether or not to shrink replaced broker ids form partition assignment')
@click.option('--broker', type=click.STRING, help='Optional broker id to execute check on')
@click.option('--parallelism', type=click.INT, show_default=True, default=1,
              help="Amount of partitions to move in a single migration step")
def migrate_broker(from_: str, to: str, shrink: bool, broker: str, parallelism: int):
    config, env_provider = prepare_configs()
    with load_exhibitor_proxy(env_provider.get_address_provider(), config.zk_prefix) as zookeeper:
        broker_id = get_opt_broker_id(broker, config, zookeeper, env_provider) if broker else None
        RemoteCommandExecutorCheck.register_migration(zookeeper, from_.split(','), to.split(','), shrink, broker_id,
                                                      parallelism)


@cli.command('swap_fat_slim', help='Move one partition from fat broker to slim one')
@click.option('--threshold', type=click.INT, default="100000", show_default=True, help="Threshold in kb to run swap")
def swap_partitions(threshold: int):
    config, env_provider = prepare_configs()
    with load_exhibitor_proxy(env_provider.get_address_provider(), config.zk_prefix) as zookeeper:
        RemoteCommandExecutorCheck.register_fatboy_slim(zookeeper, threshold_kb=threshold)


@cli.group(name='actions', help='Work with running actions')
def actions():
    pass


@actions.command('list', help='List all the actions on broker(s)')
@click.option('--broker', type=click.STRING,
              help='Broker id to list actions on. By default all brokers are enumerated')
def list_actions(broker: str):
    table = []
    config, env_provider = prepare_configs()

    for broker_id, address in _list_broker_addresses(config, env_provider, broker):
        try:
            response = requests.get('http://{}:{}/api/controller/queue'.format(address, config.health_port))
        except Exception as e:
            print('Failed to query information on {} ({})'.format(broker_id, address))
            _LOG.error('Failed to query information on {} ({})'.format(broker_id, address), exc_info=e)
            continue
        line = {
            '_broker_id': broker_id,
            '_broker_address': address,
        }
        if response.status_code != 200:
            line['error'] = _extract_error(response)
            table.append(line)
        else:
            changes = response.json()
            if not changes:
                line.update({
                    'type': None,
                    'description': None,
                    'running': None
                })
                table.append(line)
            else:
                for change in changes:
                    line_copy = dict(line)
                    line_copy.update(change)
                    table.append(line_copy)
    if not table:
        print('No brokers found')
    else:
        _print_table(table)


@actions.command('delete', help='Remove all actions of specified type on broker(s)')
@click.option('--action', type=click.STRING,
              help='Action to delete')
@click.option('--broker', type=click.STRING,
              help='Broker id to delete actions on. By default actions are deleted on all brokers')
def delete_actions(action: str, broker: str):
    if not action:
        print('No action specified. Please specify it')
    config, env_provider = prepare_configs()

    for broker_id, address in _list_broker_addresses(config, env_provider, broker):
        try:
            response = requests.delete(
                'http://{}:{}/api/controller/queue/{}'.format(address, config.health_port, action))
        except Exception as e:
            print('Failed to query information on {} ({})'.format(broker_id, address))
            _LOG.error('Failed to query information on {} ({})'.format(broker_id, address), exc_info=e)
            continue
        if response.status_code not in (200, 204):
            print('Failed to delete action from {} ({}): {}'.format(broker, address, _extract_error(response)))
        else:
            print('Removed action {} from {} ({})'.format(action, broker_id, address))


def _extract_error(response: Response):
    try:
        return response.json()['message']
    except Exception as e:
        _LOG.error('Failed to parse response message', exc_info=e)
        return response.text()


def _list_broker_addresses(config, env_provider, broker):
    with load_exhibitor_proxy(env_provider.get_address_provider(), config.zk_prefix) as zookeeper:
        for broker_id in zookeeper.get_broker_ids():
            if broker and broker != broker_id:
                continue
            yield broker_id, zookeeper.get_broker_address(broker_id)


@cli.command('stats', help='Display statistics about brokers')
def show_stats():
    config, env_provider = prepare_configs()
    with load_exhibitor_proxy(env_provider.get_address_provider(), config.zk_prefix) as zookeeper:
        disk_stats = zookeeper.get_disk_stats()
        table = []
        for broker_id in zookeeper.get_broker_ids():
            disk = disk_stats.get(broker_id, {}).get('disk') if disk_stats else {}
            table.append({
                'Broker Id': broker_id,
                'Address': zookeeper.get_broker_address(broker_id),
                'Free kb': disk.get('free_kb'),
                'Used kb': disk.get('used_kb')
            })
        _print_table(table)


@cli.group(name='validate', help='Validates internal structures of kafka/zk')
def validate():
    pass


@validate.command('replication', help='Returns all partitions whose ISR size differs from the replication factor or '
                                      'have not registered broker ids')
@click.option('--factor', type=click.INT, default=3, show_default=True, help='Replication factor')
def validate_replication(factor: int):
    config, env_provider = prepare_configs()
    with load_exhibitor_proxy(env_provider.get_address_provider(), config.zk_prefix) as zookeeper:
        brokers = {int(x) for x in zookeeper.get_broker_ids()}
        table = []
        for topic_name, partition, state in zookeeper.load_partition_states():
            if len(state['isr']) != factor or not set(state['isr']).issubset(brokers):
                table.append({
                    'Partition': partition,
                    'Topic': topic_name,
                    'State': state
                })
        if table:
            _LOG.info('Invalid topics:')
            _print_table(table)
        else:
            print('All replica lists look valid')


if __name__ == '__main__':
    cli()



import logging

from instance_control import config
from instance_control import volume
from instance_control.aws import AWSResources
from instance_control.aws.ec2_node import EC2
from instance_control.command import Command

_LOG = logging.getLogger('bubuku.cluster.command.attach')


class AttachCommand(Command):
    def __init__(self, cluster_config_path: str, volume_id: str):
        super().__init__(cluster_config_path)
        self.volume_id = volume_id

    def execute(self):
        aws_ = AWSResources(region=self.cluster_config['region'])

        vol = aws_.ec2_resource.Volume(self.volume_id)
        volume.check_volume_available(vol)
        vol.create_tags(Tags=[{'Key': 'Name', 'Value': config.KAFKA_LOGS_EBS}])
        _LOG.info('Created tag:Name=%s for %s', config.KAFKA_LOGS_EBS, self.volume_id)

        self.cluster_config['create_ebs'] = False
        self.cluster_config['availability_zone'] = vol.availability_zone

        ec2 = EC2(aws_)
        ec2.create(self.cluster_config, 1)
        volume.wait_volumes_attached(aws_)



import logging

from instance_control.aws import AWSResources

_LOG = logging.getLogger('bubuku.cluster.aws.metric')


def create_auto_recovery_alarm(aws_: AWSResources, cluster_name: str, instance_id):
    _LOG.info('Creating AWS auto recovery alarm for %s', instance_id)

    alarm_actions = ['arn:aws:automate:{}:ec2:recover'.format(aws_.region)]

    alarm_name = '{}-{}-auto-recover'.format(cluster_name, instance_id)

    aws_.cloudwatch_client.put_metric_alarm(
        AlarmName=alarm_name,
        AlarmActions=alarm_actions,
        MetricName='StatusCheckFailed_System',
        Namespace='AWS/EC2',
        Statistic='Minimum',
        Dimensions=[{
            'Name': 'InstanceId',
            'Value': instance_id
        }],
        Period=60,  # 1 minute
        EvaluationPeriods=2,
        Threshold=0,
        ComparisonOperator='GreaterThanThreshold')
    _LOG.info('Created alarm %s', alarm_name)



import logging
import threading
import time
from queue import Queue, Empty, Full

__COMMAND_QUEUE = Queue()

_LOG = logging.getLogger('bubuku.communicate')


def sleep_and_operate(controller, timeout: float):
    cur_time = time.time()
    finish = cur_time + (0.1 if timeout <= 0 else timeout)
    while cur_time < finish:
        try:
            command = __COMMAND_QUEUE.get(block=True, timeout=finish - cur_time)
            try:
                command(controller)
            except Exception as e:
                _LOG.error('Command finished with error', exc_info=e)
        except Empty:
            pass
        cur_time = time.time()


def execute_on_controller_thread(function, timeout):
    condition = threading.Condition()
    result = [None, True]

    def _execute(controller):
        with condition:
            if result[1]:
                try:
                    result[0] = function(controller)
                finally:
                    condition.notify()

    finish = time.time() + timeout
    with condition:
        try:
            __COMMAND_QUEUE.put(_execute, timeout=timeout)
        except Full:
            raise TimeoutError('Timeout expired')
        if condition.wait(timeout=finish - time.time()):
            return result[0]
        else:
            result[1] = False
            raise TimeoutError('Timeout expired')



import logging

from bubuku.aws import AWSResources

_LOG = logging.getLogger('bubuku.aws.volume')
KAFKA_LOGS_EBS = 'kafka-logs-ebs'


def are_volumes_attached(aws_: AWSResources):
    _LOG.info('Searching for volumes with tag %s to attach', KAFKA_LOGS_EBS)
    response = aws_.ec2_client.describe_volumes(Filters=[{'Name': 'tag:Name', 'Values': [KAFKA_LOGS_EBS]}])
    volumes = [aws_.ec2_resource.Volume(v['VolumeId']) for v in response['Volumes']]

    _LOG.info('Waiting for %s to be attached', volumes)
    volumes = [v for v in volumes if clear_tag(v)]
    return len(volumes) == 0


def clear_tag(v):
    v.load()
    if v.state == 'in-use':
        _LOG.info('Volume %s is attached. Clearing tag:Name', v)
        v.create_tags(Tags=[{'Key': 'Name', 'Value': ''}])
        _LOG.info('Completed clearing tag:Name for %s', v)
        return False
    return True


def check_volume_available(v):
    v.load()
    if v.state != 'available':
        _LOG.info('Volume %s is attached. Clearing tag:Name', v)
        raise Exception('Volume {} is not available for attaching. State: {}'.format(v.VolumeId, v.state))



import logging

from bubuku.aws import AWSResources
from bubuku.aws.cluster_config import ClusterConfig

_LOG = logging.getLogger('bubuku.aws.node')


def terminate(aws_: AWSResources, cluster_config: ClusterConfig, instance):
    _LOG.info('Terminating %s in %s', instance, cluster_config.get_cluster_name())
    _delete_alarm(aws_, cluster_config.get_cluster_name(), instance)
    instance.terminate()


# def is_terminated(instance):
#     instance.load()
#     _LOG.info('Instance state is %s. Waiting ...', instance.state['Name'])
#     if instance.state['Name'] == 'terminated':
#         _LOG.info('%s is successfully terminated', instance)
#         return True
#     return False


def _delete_alarm(aws_: AWSResources, cluster_name: str, instance):
    alarm_name = '{}-{}-auto-recover'.format(cluster_name, instance.instance_id)
    _LOG.info('Deleting alarm %s in %s for %s', alarm_name, cluster_name, instance)
    aws_.cloudwatch_client.delete_alarms(AlarmNames=[alarm_name])


def get_instance_by_ip(ec2_resource, cluster_config: ClusterConfig, ip):
    instances = ec2_resource.instances.filter(Filters=[
        {'Name': 'instance-state-name', 'Values': ['running', 'pending']},
        {'Name': 'network-interface.addresses.private-ip-address', 'Values': [ip]},
        {'Name': 'tag:Name', 'Values': [cluster_config.get_cluster_name()]}])
    instances = list(instances)
    if not instances:
        raise Exception('Instance by ip {} not found in cluster {}'.format(ip, cluster_config.get_cluster_name()))
    _LOG.info('Found %s by ip %s', instances[0], ip)
    return instances[0]



#!/usr/bin/env python3
"""General Kafka Start Script."""

import logging

from bubuku import health, api
from bubuku.broker import BrokerManager, StartupTimeout
from bubuku.process import KafkaProcess
from bubuku.config import load_config, KafkaProperties, Config
from bubuku.controller import Controller
from bubuku.env_provider import EnvProvider
from bubuku.features.data_size_stats import GenerateDataSizeStatistics
from bubuku.features.rebalance.check import RebalanceOnStartCheck, RebalanceOnBrokerListCheck
from bubuku.features.remote_exec import RemoteCommandExecutorCheck
from bubuku.features.restart_if_dead import CheckBrokerStopped
from bubuku.features.restart_on_zk_change import CheckExhibitorAddressChanged, RestartBrokerChange
from bubuku.features.swap_partitions import CheckBrokersDiskImbalance
from bubuku.features.terminate import register_terminate_on_interrupt
from bubuku.utils import CmdHelper
from bubuku.zookeeper import BukuExhibitor, load_exhibitor_proxy

_LOG = logging.getLogger('bubuku.main')


def apply_features(api_port, features: dict, controller: Controller, buku_proxy: BukuExhibitor, broker: BrokerManager,
                   kafka_properties: KafkaProperties, env_provider: EnvProvider) -> list:
    for feature, config in features.items():
        if feature == 'restart_on_exhibitor':
            controller.add_check(CheckExhibitorAddressChanged(buku_proxy, broker))
        elif feature == 'rebalance_on_start':
            controller.add_check(RebalanceOnStartCheck(buku_proxy, broker))
        elif feature == 'rebalance_on_brokers_change':
            controller.add_check(RebalanceOnBrokerListCheck(buku_proxy, broker))
        elif feature == 'balance_data_size':
            controller.add_check(
                CheckBrokersDiskImbalance(buku_proxy, broker, config["diff_threshold_mb"] * 1024, api_port))
        elif feature == 'graceful_terminate':
            register_terminate_on_interrupt(controller, broker)
        elif feature == 'use_ip_address':
            kafka_properties.set_property('advertised.host.name', env_provider.get_id())
        else:
            _LOG.error('Using of unsupported feature "{}", skipping it'.format(feature))


def run_daemon_loop(config: Config, process_holder: KafkaProcess, cmd_helper: CmdHelper, restart_on_init: bool):
    _LOG.info("Using configuration: {}".format(config))
    kafka_props = KafkaProperties(config.kafka_settings_template,
                                  '{}/config/server.properties'.format(config.kafka_dir))

    env_provider = EnvProvider.create_env_provider(config)
    address_provider = env_provider.get_address_provider()
    rack = env_provider.get_rack()
    if rack:
        kafka_props.set_property('broker.rack', rack)
    startup_timeout = StartupTimeout.build(config.timeout)

    _LOG.info("Loading exhibitor configuration")
    with load_exhibitor_proxy(address_provider, config.zk_prefix) as zookeeper:
        _LOG.info("Loading broker_id policy")
        broker_id_manager = env_provider.create_broker_id_manager(zookeeper, kafka_props)

        _LOG.info("Building broker manager")
        broker = BrokerManager(process_holder, zookeeper, broker_id_manager, kafka_props,
                               startup_timeout)

        _LOG.info("Creating controller")
        controller = Controller(broker, zookeeper, env_provider)

        controller.add_check(CheckBrokerStopped(broker, zookeeper))
        controller.add_check(RemoteCommandExecutorCheck(zookeeper, broker, config.health_port))
        controller.add_check(GenerateDataSizeStatistics(zookeeper, broker, cmd_helper,
                                                        kafka_props.get_property("log.dirs").split(",")))
        apply_features(config.health_port, config.features, controller, zookeeper, broker, kafka_props, env_provider)

        _LOG.info('Starting main controller loop')
        controller.loop(RestartBrokerChange(zookeeper, broker, lambda: False) if restart_on_init else None)


def main():
    logging.basicConfig(level=getattr(logging, 'INFO', None))

    config = load_config()
    _LOG.info("Using configuration: {}".format(config))
    process = KafkaProcess(config.kafka_dir)
    _LOG.info('Starting health server')
    cmd_helper = CmdHelper()
    health.start_server(config.health_port, cmd_helper)
    api.start_api_server()
    restart_on_init = False
    while True:
        try:
            run_daemon_loop(config, process, cmd_helper, restart_on_init)
            break
        except Exception as ex:
            _LOG.error("WOW! Almost died! Will try to restart from the begin. "
                       "After initialization will be complete, will try to restart", exc_info=ex)
            if process.is_running():
                restart_on_init = False


if __name__ == '__main__':
    main()



from bubuku.controller import Change


class BaseRebalanceChange(Change):
    def get_name(self) -> str:
        return 'rebalance'

    def can_run(self, current_actions):
        return all([a not in current_actions for a in ['start', 'restart', 'rebalance', 'stop']])

    @staticmethod
    def should_be_paused(current_actions):
        return any([a in current_actions for a in ['restart', 'start', 'stop']])


import logging

_LOG = logging.getLogger('bubuku.cluster.aws.taupage')


def find_amis(ec2_resource, region: str) -> dict:
    '''
    Find latest Taupage AMI for the region
    '''
    _LOG.info('Finding latest Taupage AMI in %s..', region)

    filters = [{'Name': 'name', 'Values': ['*Taupage-AMI-*']},
               {'Name': 'is-public', 'Values': ['false']},
               {'Name': 'state', 'Values': ['available']},
               {'Name': 'root-device-type', 'Values': ['ebs']}]
    images = list(ec2_resource.images.filter(Filters=filters))
    if not images:
        raise Exception('No Taupage AMI found')
    most_recent_image = sorted(images, key=lambda i: i.name)[-1]

    _LOG.info('The latest AMI is %s', most_recent_image)

    return most_recent_image



import logging

import netaddr

from bubuku.aws.cluster_config import ClusterConfig
from bubuku.aws import AWSResources

_LOG = logging.getLogger('bubuku.cluster.aws.subnet')


def get_subnets(ec2, prefix_filter: str, cluster_config: ClusterConfig) -> list:
    '''
    Returns lists of subnets, which names start
    with the specified prefix (it should be either 'dmz-' or
    'internal-'), sorted by the Availability Zone and filtered by vpc id
    '''
    vpc_id = cluster_config.get_vpc_id()
    availability_zone = cluster_config.get_availability_zone()
    _LOG.info('Getting subnets for vpc_id: %s and availability_zone: %s', vpc_id, availability_zone)

    resp = ec2.describe_subnets()
    subnets = []

    for subnet in resp['Subnets']:
        if subnet['VpcId'] != vpc_id:
            continue
        for tag in subnet['Tags']:
            if tag['Key'] == 'Name' and tag['Value'].startswith(prefix_filter):
                if availability_zone:
                    if subnet['AvailabilityZone'] == availability_zone:
                        subnets.append(subnet)
                else:
                    subnets.append(subnet)
    _LOG.info('Got subnets %s ', subnets)
    return subnets


class IpAddressPoolDepletedException(Exception):
    def __init__(self, cidr_block: str):
        msg = "Pool of unused IP addresses depleted in subnet: {}".format(cidr_block)
        super(IpAddressPoolDepletedException, self).__init__(msg)


def allocate_ip_addresses(aws_: AWSResources, cluster_config, address_count: int) -> list:
    '''
    Allocate unused private IP addresses by checking the current
    reservations
    Return list of tuples (subnet, ip)
    '''
    _LOG.info('Allocating IP addresses ...')

    def try_next_address(ips, subnet):
        try:
            return str(next(ips))
        except StopIteration:
            raise IpAddressPoolDepletedException(subnet['CidrBlock'])

    #
    # Here we have to account for the behavior of launch_*_nodes
    # which iterate through subnets to put the instances into
    # different Availability Zones.
    #
    subnets = get_subnets(aws_.ec2_client, 'internal-', cluster_config)
    network_ips = [netaddr.IPNetwork(s['CidrBlock']).iter_hosts() for s in subnets]

    for idx, ips in enumerate(network_ips):
        #
        # Some of the first addresses in each subnet are
        # taken by AWS system instances that we can't see,
        # so we try to skip them.
        #
        for _ in range(10):
            try_next_address(ips, subnets[idx])

    i = 0
    result_ips = []
    while i < address_count:
        idx = i % len(subnets)
        subnet = subnets[idx]
        ip = try_next_address(network_ips[idx], subnet)
        resp = aws_.ec2_client.describe_instances(Filters=[{
            'Name': 'private-ip-address',
            'Values': [ip]
        }])
        if not resp['Reservations']:
            i += 1
            _LOG.info('Got ip address %s ', ip)
            result_ips.append((subnet, ip))

    _LOG.info('IP Addresses are allocated')

    return result_ips



import logging

from bubuku.aws import AWSResources
from bubuku.aws.cluster_config import ClusterConfig

_LOG = logging.getLogger('bubuku.aws.security_group')


def create_or_ger_security_group(aws_: AWSResources, cluster_config: ClusterConfig) -> dict:
    _LOG.info('Configuring security group ...')
    security_groups = aws_.ec2_client.describe_security_groups(
        Filters=[{'Name': 'group-name', 'Values': [cluster_config.get_cluster_name()]}])
    if security_groups['SecurityGroups']:
        sg = security_groups['SecurityGroups'][0]
        _LOG.info('Security group for %s exists, will use it %s', cluster_config.get_cluster_name(), sg['GroupId'])
        return sg

    _LOG.info("Security group does not exists, creating ...")
    sg_name = cluster_config.get_cluster_name()
    sg = aws_.ec2_client.create_security_group(
        GroupName=sg_name,
        VpcId=cluster_config.get_vpc_id(),
        Description='Bubuku Security Group')
    _LOG.info("Security group %s is created", sg_name)
    aws_.ec2_client.create_tags(Resources=[sg['GroupId']], Tags=[{'Key': 'Name', 'Value': sg_name}])
    aws_.ec2_client.authorize_security_group_ingress(
        GroupId=sg['GroupId'],
        IpPermissions=[get_ip_permission(22), get_ip_permission(8004),
                       get_ip_permission(8778), get_ip_permission(9100),
                       get_ip_permission(9092),
                       get_ip_permission(cluster_config.get_health_port())])
    _LOG.info("Security group got ingress for ports: 22, 8004, 8080, 8778, 9100, %s", cluster_config.get_health_port())
    return sg


def get_ip_permission(port: int):
    return {
        'IpProtocol': 'tcp',
        'FromPort': port,
        'ToPort': port,
        'IpRanges': [{'CidrIp': '0.0.0.0/0'}]
    }



import json
import logging
import re

import requests

_LOG = logging.getLogger('bubuku.cluster.config')
KAFKA_LOGS_EBS = 'kafka-logs-ebs'


def read_cluster_config(cluster_config_path: str):
    _LOG.info('Reading config from %s', cluster_config_path)
    with open(cluster_config_path) as data_file:
        return json.load(data_file)


def validate_config(cluster_config: dict):
    _LOG.info("Validating configuration for %s", cluster_config['cluster_name'])
    cluster_name_re = '^[a-z][a-z0-9-]*[a-z0-9]$'
    if not re.match(cluster_name_re, cluster_config['cluster_name']):
        raise Exception(
            'Cluster name must only contain lowercase latin letters, digits and dashes '
            '(it also must start with a letter and cannot end with a dash), in other words '
            'it must be matched by the following regular expression: {}'.format(cluster_name_re))

    if not cluster_config['region']:
        raise Exception('Please specify at least one region')

    artifact_name = 'bubuku-appliance'
    url = 'https://registry.opensource.zalan.do/teams/aruha/artifacts/{}/tags'.format(artifact_name)
    if not next((tag for tag in requests.get(url).json() if tag['name'] == cluster_config['image_version']), None):
        raise Exception('Docker image was not found')

    _LOG.info(json.dumps(cluster_config, indent=4))



import logging
import signal

from bubuku.broker import BrokerManager
from bubuku.controller import Controller, Change

_LOG = logging.getLogger('bubuku.features.terminate')


class StopBrokerChange(Change):
    def __init__(self, broker: BrokerManager):
        self.broker = broker

    def get_name(self):
        return 'stop'

    def __str__(self):
        return 'StopBrokerChange ({})'.format(self.get_name())

    def can_run(self, current_actions):
        return all([action not in current_actions for action in ['start', 'restart', 'stop']])

    def run(self, current_actions):
        _LOG.info('Stopping kafka process')
        self.broker.stop_kafka_process()
        return self.broker.has_leadership()

    def can_run_at_exit(self):
        return True


__REGISTERED = None


def get_registration():
    if not __REGISTERED:
        return None, None
    return __REGISTERED


def register_terminate_on_interrupt(controller: Controller, broker: BrokerManager):
    global __REGISTERED

    def _sig_handler(*args, **kwargs):
        _LOG.info('Signal was caught, stopping controller gracefully')
        controller.stop(StopBrokerChange(broker))

    _LOG.info('Registering signal handler')
    old_handler = signal.signal(signal.SIGTERM, _sig_handler)
    if old_handler:
        _LOG.warn('Old handler is removed: {}'.format(old_handler))
    __REGISTERED = (controller, broker)



import logging

from instance_control import volume
from instance_control.aws import AWSResources
from instance_control.aws.ec2_node import EC2
from instance_control.command import Command

_LOG = logging.getLogger('bubuku.cluster.command.attach')


class CreateCommand(Command):
    def __init__(self, cluster_config_path: str,
                 count: int,
                 availability_zone: str,
                 image_version: str):
        super().__init__(cluster_config_path)
        self.count = count
        self.availability_zone = availability_zone
        self.image_version = image_version

    def alter_config(self):
        if self.image_version:
            self.cluster_config['image_version'] = self.image_version
        self.cluster_config['create_ebs'] = True
        self.cluster_config['availability_zone'] = self.availability_zone

    def execute(self):
        aws_ = AWSResources(region=self.cluster_config['region'])
        ec2 = EC2(aws_)

        ec2.create(self.cluster_config, self.count)
        volume.wait_volumes_attached(aws_)



import logging
from collections import namedtuple
from operator import attrgetter
from typing import List

import requests

from bubuku.broker import BrokerManager
from bubuku.controller import Check
from bubuku.features.rebalance import BaseRebalanceChange
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.features.swap_partitions')

TpData = namedtuple('_TpData', ('topic', 'partition', 'size', 'replicas'))


class SwapPartitionsChange(BaseRebalanceChange):
    def __init__(self, zk: BukuExhibitor, swap_data_provider):
        self.zk = zk
        self.to_move = None
        self.swap_data_provider = swap_data_provider

    def run(self, current_actions):
        if self.should_be_paused(current_actions):
            _LOG.info("Pausing swap partitions change as there are conflicting actions: {}".format(current_actions))
            return True
        # if there's a rebalance currently running - postpone current change
        if self.zk.is_rebalancing():
            return True

        if self.to_move is None:
            slim_broker_id, fat_broker_id, gap, size_stats = self.swap_data_provider(self.zk)
            if slim_broker_id is None:
                _LOG.info('Can not find slim broker and fat broker during reassignment. Probably gap changed')
                return False
            # merge topics size stats to a single dict
            topics_stats = {}
            for broker_stats in size_stats.values():
                for topic in broker_stats["topics"].keys():
                    if topic not in topics_stats:
                        topics_stats[topic] = {}
                    topics_stats[topic].update(broker_stats["topics"][topic])

            # find partitions that are candidates to be swapped between "fat" and "slim" brokers
            swap_partition_candidates = self.__find_all_swap_candidates(fat_broker_id, slim_broker_id, topics_stats)

            # smallest partition from slim broker is the one we move to fat broker
            slim_broker_smallest_partition = min(swap_partition_candidates[slim_broker_id], key=attrgetter("size"))
            if not slim_broker_smallest_partition:
                _LOG.info("No partitions on slim broker(id: {}) found to swap".format(slim_broker_id))
                return False
            _LOG.info("Slim broker(id: {}) partition to swap: {}".format(
                slim_broker_id, slim_broker_smallest_partition))

            # find the best fitting fat broker partition to move to slim broker
            # (should be as much as possible closing the gap between brokers)
            fat_broker_swap_candidates = swap_partition_candidates[fat_broker_id]
            matching_swap_partition = self.__find_best_swap_candidate(fat_broker_swap_candidates, gap,
                                                                      slim_broker_smallest_partition.size)

            # if there is no possible swap that will decrease the gap - just do nothing
            if not matching_swap_partition:
                _LOG.info("No candidate from fat broker(id:{}) found to swap".format(fat_broker_id))
                return False
            _LOG.info("Fat broker(id: {}) partition to swap: {}".format(fat_broker_id, matching_swap_partition))
            # write rebalance-json to ZK; Kafka will read it and perform the partitions swap
            self.to_move = self.__create_rebalance_list(slim_broker_smallest_partition, slim_broker_id,
                                                        matching_swap_partition, fat_broker_id)

        # if there is already a swap which was postponed - just execute it
        return not self.__perform_swap(self.to_move)

    def __perform_swap(self, rebalance_list):
        _LOG.info("Writing rebalance-json to ZK for partitions swap: {}".format(rebalance_list))
        return self.zk.reallocate_partitions(rebalance_list)

    def __find_all_swap_candidates(self, fat_broker_id: int, slim_broker_id: int, topics_stats: dict) -> dict:
        swap_partition_candidates = {fat_broker_id: [], slim_broker_id: []}
        for topic, partition, replicas in self.zk.load_partition_assignment():
            if topic not in topics_stats or str(partition) not in topics_stats[topic]:
                continue  # we skip this partition as there is not data size stats for it

            if replicas[0] in (fat_broker_id, slim_broker_id):
                continue  # Skip leadership transfer

            if fat_broker_id in replicas and slim_broker_id in replicas:
                continue  # we skip this partition as it exists on both involved brokers

            for broker_id in [slim_broker_id, fat_broker_id]:
                if broker_id in replicas:
                    swap_partition_candidates[broker_id].append(
                        TpData(topic, partition, topics_stats[topic][str(partition)], replicas))
        return swap_partition_candidates

    @staticmethod
    def __find_best_swap_candidate(candidates: list, brokers_gap: int, partition_to_swap_size: int) -> TpData:
        candidates.sort(key=attrgetter("size"), reverse=True)
        matching_swap_partition = None
        smallest_new_gap = brokers_gap
        for tp in candidates:
            new_gap = abs(brokers_gap - 2 * abs(tp.size - partition_to_swap_size))
            if new_gap < smallest_new_gap:
                smallest_new_gap = new_gap
                matching_swap_partition = tp
        return matching_swap_partition

    def __create_rebalance_list(self, tp1: TpData, br1: int, tp2: TpData, br2: int) -> list:
        return [
            (tp1.topic, tp1.partition, self.__replace_broker(tp1.replicas, br1, br2, tp2.replicas[0] == br2)),
            (tp2.topic, tp2.partition, self.__replace_broker(tp2.replicas, br2, br1, tp1.replicas[0] == br1))
        ]

    def __replace_broker(self, replicas: list, broker_to_replace: int, replacement: int, was_leader: bool) -> list:
        rps = [x for x in replicas if x != broker_to_replace]
        if was_leader:
            return [replacement] + rps
        else:
            return rps + [replacement]

    def __str__(self):
        return 'SwapPartitions'


def _load_disk_stats(zk: BukuExhibitor, api_port: int):
    size_stats = zk.get_disk_stats()
    if len(size_stats) < 2:
        _LOG.info("No size stats available, imbalance check cancelled")
        return None
    result = {}
    for broker_id, value in size_stats.items():
        try:
            if api_port != -1:  # For unit tests only
                host = zk.get_broker_address(broker_id)
                tmp = requests.get(
                    'http://{}:{}/api/disk_stats'.format(host, api_port),
                    timeout=5).json()
                if any(a not in tmp for a in ['free_kb', 'used_kb']):
                    continue
                value['disk'] = tmp
                value['host'] = host
            result[broker_id] = value
        except Exception as e:
            _LOG.error('Failed to load disk stats for broker {}. Skipping it'.format(broker_id), exc_info=e)

    return result


def load_swap_data(zk: BukuExhibitor, api_port: int, gap: int) -> (str, str, int, dict):
    """
    Finds brokers that could be used for gap of size gap. If rack awareness is enabled, the swap will be between two
    brokers in the same rack
    :param zk: Bubuku exhibitor
    :param api_port: bubuku api port
    :param gap: gap in kb to get information for
    :return: (slim_broker_id, fat_broker_id, calculated_gap, size_stats) or (None, None, calculated_gap, size_stats)
    """
    size_stats = _load_disk_stats(zk, api_port)
    if not size_stats or len(size_stats) < 2:
        return None, None, None, size_stats
    sorted_stats = sorted(size_stats.items(), key=lambda tup: tup[1]["disk"]["free_kb"])
    fat_broker, slim_broker = select_fat_slim_brokers(zk, sorted_stats)
    if fat_broker is None:
        return None, None, None, size_stats

    calculated_gap = slim_broker[1]['disk']['free_kb'] - fat_broker[1]['disk']['free_kb']
    _LOG.info('Gap between {} and {} is {}, need to fix: {}'.format(
        fat_broker[0], slim_broker[0], calculated_gap, calculated_gap > gap))
    if calculated_gap >= gap:
        return int(slim_broker[0]), int(fat_broker[0]), calculated_gap, size_stats
    return None, None, calculated_gap, size_stats


def select_fat_slim_brokers(zk: BukuExhibitor, sorted_stats: list):
    racks = zk.get_broker_racks()
    if any([rack is None for rack in racks.values()]):
        return sorted_stats[0], sorted_stats[-1]
    for i in range(len(sorted_stats) - 1):
        fat_broker = sorted_stats[i]
        fat_rack = racks[int(fat_broker[0])]
        for j in range(len(sorted_stats) -1, i, -1):
            slim_broker = sorted_stats[j]
            slim_rack = racks[int(slim_broker[0])]
            if slim_rack == fat_rack:
                return fat_broker, slim_broker

    return None, None


class CheckBrokersDiskImbalance(Check):
    def __init__(self, zk: BukuExhibitor, broker: BrokerManager, diff_threshold_kb: int, api_port: int):
        super().__init__(check_interval_s=900)
        self.zk = zk
        self.api_port = api_port
        self.broker = broker
        self.diff_threshold_kb = diff_threshold_kb

    def check(self):
        if self.broker.is_running_and_registered():
            _LOG.info("Starting broker disk imbalance check")
            try:
                slim_broker_id, fat_broker_id, gap, size_stats = load_swap_data(
                    self.zk, self.api_port, self.diff_threshold_kb)
                if slim_broker_id is not None:  # All or nothing
                    return SwapPartitionsChange(
                        self.zk,
                        lambda x: load_swap_data(x, self.api_port, self.diff_threshold_kb))
            except Exception as e:
                _LOG.warn("Error occurred when performing disk imbalance check", exc_info=e)
        return None

    def __str__(self):
        return 'CheckBrokersDiskImbalance'
